project = "uritools"
copyright = "2014-2023 Thomas Kemmer"
version = "4.0"
release = "4.0.1"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.todo",
]
exclude_patterns = ["_build"]
master_doc = "index"
html_theme = "default"
